@@ -1,12 +1,13 @@
 # coding=utf-8
 #
-# This file is part of Hypothesis (https://github.com/DRMacIver/hypothesis)
+# This file is part of Hypothesis, which may be found at
+# https://github.com/HypothesisWorks/hypothesis-python
 #
-# Most of this work is copyright (C) 2013-2015 David R. MacIver
+# Most of this work is copyright (C) 2013-2016 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
-# https://github.com/DRMacIver/hypothesis/blob/master/CONTRIBUTING.rst for a
-# full list of people who may hold copyright, and consult the git log if you
-# need to determine who owns an individual contribution.
+# CONTRIBUTING.rst for a full list of people who may hold copyright, and
+# consult the git log if you need to determine who owns an individual
+# contribution.
 #
 # This Source Code Form is subject to the terms of the Mozilla Public License,
 # v. 2.0. If a copy of the MPL was not distributed with this file, You can
@@ -23,21 +24,25 @@ failing examples it finds.
 """
 
 
-from hypothesis._settings import settings, Verbosity
+from hypothesis._settings import settings, Verbosity, Phase, HealthCheck
 from hypothesis.version import __version_info__, __version__
-from hypothesis.control import assume, note
+from hypothesis.control import assume, note, reject, event
 from hypothesis.core import given, find, example, seed
 
 
 __all__ = [
     'settings',
     'Verbosity',
+    'HealthCheck',
+    'Phase',
     'assume',
+    'reject',
     'seed',
     'given',
     'find',
     'example',
     'note',
+    'event',
     '__version__',
     '__version_info__',
 ]
