@@ -3,7 +3,7 @@
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis-python
 #
-# Most of this work is copyright (C) 2013-2015 David R. MacIver
+# Most of this work is copyright (C) 2013-2016 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -21,7 +21,7 @@ from enum import IntEnum
 
 from hypothesis.errors import Frozen, InvalidArgument
 from hypothesis.internal.compat import hbytes, text_type, int_to_bytes, \
-    unicode_safe_repr, reasonable_byte_type
+    benchmark_time, unicode_safe_repr, reasonable_byte_type
 
 
 def uniform(random, n):
@@ -72,6 +72,8 @@ class TestData(object):
         global global_test_counter
         self.testcounter = global_test_counter
         global_test_counter += 1
+        self.start_time = benchmark_time()
+        self.events = set()
 
     def __assert_not_frozen(self, name):
         if self.frozen:
@@ -119,11 +121,15 @@ class TestData(object):
             if not self.intervals or self.intervals[-1] != t:
                 self.intervals.append(t)
 
+    def note_event(self, event):
+        self.events.add(event)
+
     def freeze(self):
         if self.frozen:
             assert isinstance(self.buffer, hbytes)
             return
         self.frozen = True
+        self.finish_time = benchmark_time()
         # Intervals are sorted as longest first, then by interval start.
         for l in self.intervals_by_level:
             for i in range(len(l) - 1):
@@ -134,6 +140,7 @@ class TestData(object):
             key=lambda se: (se[0] - se[1], se[0])
         )
         self.buffer = hbytes(self.buffer)
+        self.events = frozenset(self.events)
         del self._draw_bytes
 
     def draw_bytes(self, n, distribution=uniform):
