@@ -3,7 +3,7 @@
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis-python
 #
-# Most of this work is copyright (C) 2013-2015 David R. MacIver
+# Most of this work is copyright (C) 2013-2016 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -18,7 +18,9 @@
 from __future__ import division, print_function, absolute_import
 
 import time
+from enum import Enum
 from random import Random, getrandbits
+from weakref import WeakKeyDictionary
 
 from hypothesis import settings as Settings
 from hypothesis import Phase
@@ -29,6 +31,15 @@ from hypothesis.internal.conjecture.data import Status, StopTest, TestData
 from hypothesis.internal.conjecture.minimizer import minimize
 
 
+class ExitReason(Enum):
+    max_examples = 0
+    max_iterations = 1
+    timeout = 2
+    max_shrinks = 3
+    finished = 4
+    flaky = 5
+
+
 class RunIsComplete(Exception):
     pass
 
@@ -44,13 +55,16 @@ class TestRunner(object):
         self.last_data = None
         self.changed = 0
         self.shrinks = 0
-        self.examples_considered = 0
-        self.iterations = 0
+        self.call_count = 0
+        self.event_call_counts = Counter()
         self.valid_examples = 0
         self.start_time = time.time()
         self.random = random or Random(getrandbits(128))
         self.database_key = database_key
         self.seen = set()
+        self.duplicates = 0
+        self.status_runtimes = {}
+        self.events_to_strings = WeakKeyDictionary()
 
     def new_buffer(self):
         self.last_data = TestData(
@@ -60,10 +74,9 @@ class TestRunner(object):
         )
         self.test_function(self.last_data)
         self.last_data.freeze()
-        self.note_for_corpus(self.last_data)
 
     def test_function(self, data):
-        self.iterations += 1
+        self.call_count += 1
         try:
             self._test_function(data)
             data.freeze()
@@ -74,6 +87,9 @@ class TestRunner(object):
         except:
             self.save_buffer(data.buffer)
             raise
+        finally:
+            data.freeze()
+            self.note_details(data)
         if (
             data.status == Status.INTERESTING and (
                 self.last_data is None or
@@ -90,7 +106,11 @@ class TestRunner(object):
         #   2. Any transition which increases the status is valid
         #   3. If the previous status was interesting, only shrinking
         #      transitions are allowed.
-        self.seen.add(hbytes(data.buffer))
+        key = hbytes(data.buffer)
+        if key in self.seen:
+            self.duplicates += 1
+            return False
+        self.seen.add(key)
         if data.buffer == self.last_data.buffer:
             return False
         if self.last_data.status < data.status:
@@ -118,9 +138,13 @@ class TestRunner(object):
                 self.database_key, hbytes(buffer)
             )
 
-    def note_for_corpus(self, data):
+    def note_details(self, data):
         if data.status == Status.INTERESTING:
             self.save_buffer(data.buffer)
+        runtime = max(data.finish_time - data.start_time, 0.0)
+        self.status_runtimes.setdefault(data.status, []).append(runtime)
+        for event in set(map(self.event_to_string, data.events)):
+            self.event_call_counts[event] += 1
 
     def debug(self, message):
         with self.settings:
@@ -142,20 +166,19 @@ class TestRunner(object):
             self.settings.timeout > 0 and
             time.time() >= self.start_time + self.settings.timeout
         ):
+            self.exit_reason = ExitReason.timeout
             raise RunIsComplete()
-        self.examples_considered += 1
         buffer = buffer[:self.last_data.index]
         if sort_key(buffer) >= sort_key(self.last_data.buffer):
             return False
         assert sort_key(buffer) <= sort_key(self.last_data.buffer)
         data = TestData.for_buffer(buffer)
         self.test_function(data)
-        data.freeze()
-        self.note_for_corpus(data)
         if self.consider_new_test_data(data):
             self.shrinks += 1
             self.last_data = data
             if self.shrinks >= self.settings.max_shrinks:
+                self.exit_reason = ExitReason.max_shrinks
                 raise RunIsComplete()
             self.last_data = data
             self.changed += 1
@@ -170,7 +193,7 @@ class TestRunner(object):
                 pass
             self.debug(
                 u'Run complete after %d examples (%d valid) and %d shrinks' % (
-                    self.iterations, self.valid_examples, self.shrinks,
+                    self.call_count, self.valid_examples, self.shrinks,
                 ))
 
     def _new_mutator(self):
@@ -253,10 +276,12 @@ class TestRunner(object):
             )
             for existing in corpus:
                 if self.valid_examples >= self.settings.max_examples:
+                    self.exit_reason = ExitReason.max_examples
                     return
-                if self.iterations >= max(
+                if self.call_count >= max(
                     self.settings.max_iterations, self.settings.max_examples
                 ):
+                    self.exit_reason = ExitReason.max_iterations
                     return
                 data = TestData.for_buffer(existing)
                 self.test_function(data)
@@ -289,15 +314,18 @@ class TestRunner(object):
             mutator = self._new_mutator()
             while self.last_data.status != Status.INTERESTING:
                 if self.valid_examples >= self.settings.max_examples:
+                    self.exit_reason = ExitReason.max_examples
                     return
-                if self.iterations >= max(
+                if self.call_count >= max(
                     self.settings.max_iterations, self.settings.max_examples
                 ):
+                    self.exit_reason = ExitReason.max_iterations
                     return
                 if (
                     self.settings.timeout > 0 and
                     time.time() >= start_time + self.settings.timeout
                 ):
+                    self.exit_reason = ExitReason.timeout
                     return
                 if mutations >= self.settings.max_mutations:
                     mutations = 0
@@ -310,7 +338,6 @@ class TestRunner(object):
                     )
                     self.test_function(data)
                     data.freeze()
-                    self.note_for_corpus(data)
                     prev_data = self.last_data
                     if self.consider_new_test_data(data):
                         self.last_data = data
@@ -323,21 +350,26 @@ class TestRunner(object):
 
         data = self.last_data
         if data is None:
+            self.exit_reason = ExitReason.finished
             return
         assert isinstance(data.output, text_type)
 
         if self.settings.max_shrinks <= 0:
+            self.exit_reason = ExitReason.max_shrinks
             return
 
         if Phase.shrink not in self.settings.phases:
+            self.exit_reason = ExitReason.finished
             return
 
         if not self.last_data.buffer:
+            self.exit_reason = ExitReason.finished
             return
 
         data = TestData.for_buffer(self.last_data.buffer)
         self.test_function(data)
         if data.status != Status.INTERESTING:
+            self.exit_reason = ExitReason.flaky
             return
 
         change_counter = -1
@@ -451,6 +483,18 @@ class TestRunner(object):
                             alternatives = None
                             break
                 i += 1
+        self.exit_reason = ExitReason.finished
+
+    def event_to_string(self, event):
+        if isinstance(event, str):
+            return event
+        try:
+            return self.events_to_strings[event]
+        except KeyError:
+            pass
+        result = str(event)
+        self.events_to_strings[event] = result
+        return result
 
 
 def _draw_predecessor(rnd, xs):
