@@ -1,12 +1,13 @@
 # coding=utf-8
 #
-# This file is part of Hypothesis (https://github.com/DRMacIver/hypothesis)
+# This file is part of Hypothesis, which may be found at
+# https://github.com/HypothesisWorks/hypothesis-python
 #
-# Most of this work is copyright (C) 2013-2015 David R. MacIver
+# Most of this work is copyright (C) 2013-2016 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
-# https://github.com/DRMacIver/hypothesis/blob/master/CONTRIBUTING.rst for a
-# full list of people who may hold copyright, and consult the git log if you
-# need to determine who owns an individual contribution.
+# CONTRIBUTING.rst for a full list of people who may hold copyright, and
+# consult the git log if you need to determine who owns an individual
+# contribution.
 #
 # This Source Code Form is subject to the terms of the Mozilla Public License,
 # v. 2.0. If a copy of the MPL was not distributed with this file, You can
@@ -28,23 +29,24 @@ from __future__ import division, print_function, absolute_import
 
 import inspect
 import traceback
-from random import Random
 from unittest import TestCase
 from collections import namedtuple
 
+import hypothesis.internal.conjecture.utils as cu
 from hypothesis.core import find
 from hypothesis.errors import Flaky, NoSuchExample, InvalidDefinition, \
-    UnsatisfiedAssumption
+    HypothesisException
 from hypothesis.control import BuildContext
-from hypothesis.settings import Settings, Verbosity, note_deprecation
+from hypothesis._settings import settings as Settings
+from hypothesis._settings import Verbosity
 from hypothesis.reporting import report, verbose_report, current_verbosity
-from hypothesis.internal.compat import hrange, integer_types
-from hypothesis.internal.reflection import proxies
-from hypothesis.searchstrategy.misc import JustStrategy, \
-    SampledFromStrategy
-from hypothesis.internal.strategymethod import strategy
-from hypothesis.searchstrategy.strategies import BadData, check_length, \
-    SearchStrategy, check_data_type, one_of_strategies
+from hypothesis.strategies import just, lists, builds, one_of, runner, \
+    integers
+from hypothesis.vendor.pretty import CUnicodeIO, RepresentationPrinter
+from hypothesis.internal.reflection import proxies, nicerepr
+from hypothesis.internal.conjecture.data import StopTest
+from hypothesis.internal.conjecture.utils import integer_range
+from hypothesis.searchstrategy.strategies import SearchStrategy
 from hypothesis.searchstrategy.collections import TupleStrategy, \
     FixedKeysDictStrategy
 
@@ -68,7 +70,7 @@ def find_breaking_runner(state_machine_factory, settings=None):
         try:
             runner.run(state_machine_factory())
             return False
-        except (InvalidDefinition, UnsatisfiedAssumption):
+        except HypothesisException:
             raise
         except Exception:
             verbose_report(traceback.format_exc)
@@ -80,19 +82,12 @@ def find_breaking_runner(state_machine_factory, settings=None):
             settings = Settings.default
 
     search_strategy = StateMachineSearchStrategy(settings)
-    if settings.database is not None:
-        storage = settings.database.storage(
-            getattr(
-                state_machine_factory, u'__name__',
-                type(state_machine_factory).__name__))
-    else:
-        storage = None
 
     return find(
         search_strategy,
         is_breaking_run,
         settings=settings,
-        storage=storage,
+        database_key=state_machine_factory.__name__.encode('utf-8')
     )
 
 
@@ -109,10 +104,13 @@ def run_state_machine_as_test(state_machine_factory, settings=None):
         breaker = find_breaking_runner(state_machine_factory, settings)
     except NoSuchExample:
         return
-    with BuildContext(is_final=True):
-        breaker.run(state_machine_factory(), print_steps=True)
+    try:
+        with BuildContext(None, is_final=True):
+            breaker.run(state_machine_factory(), print_steps=True)
+    except StopTest:
+        pass
     raise Flaky(
-        u'Run failed initially by succeeded on a second try'
+        u'Run failed initially but succeeded on a second try'
     )
 
 
@@ -151,7 +149,7 @@ class GenericStateMachine(object):
 
         """
         self.step_count = getattr(self, u'step_count', 0) + 1
-        report(u'Step #%d: %s' % (self.step_count, repr(step)))
+        report(u'Step #%d: %s' % (self.step_count, nicerepr(step)))
 
     def teardown(self):
         """Called after a run has finished executing to clean up any necessary
@@ -174,11 +172,17 @@ class GenericStateMachine(object):
             pass
 
         class StateMachineTestCase(TestCase):
-            settings = Settings()
+            settings = Settings(
+                min_satisfying_examples=1
+            )
 
-            def runTest(self):
-                run_state_machine_as_test(state_machine_class)
+        # We define this outside of the class and assign it because you can't
+        # assign attributes to instance method values in Python 2
+        def runTest(self):
+            run_state_machine_as_test(state_machine_class)
 
+        runTest.is_hypothesis_test = True
+        StateMachineTestCase.runTest = runTest
         base_name = state_machine_class.__name__
         StateMachineTestCase.__name__ = str(
             base_name + u'.TestCase'
@@ -195,19 +199,6 @@ class GenericStateMachine(object):
 GenericStateMachine.find_breaking_runner = classmethod(find_breaking_runner)
 
 
-def seeds(starting, n_steps):
-    random = Random(starting)
-
-    result = []
-    for _ in hrange(n_steps):
-        result.append(random.getrandbits(64))
-    return result
-
-
-# Sentinel value used to mark entries as deleted.
-TOMBSTONE = [object(), [u'TOMBSTONE FOR STATEFUL TESTING']]
-
-
 class StateMachineRunner(object):
 
     """A StateMachineRunner is a description of how to run a state machine.
@@ -216,101 +207,34 @@ class StateMachineRunner(object):
 
     """
 
-    def __init__(
-        self, parameter_seed, template_seed, n_steps,
-        record=None, templates=None,
-    ):
-        self.parameter_seed = parameter_seed
-        self.template_seed = template_seed
+    def __init__(self, data, n_steps):
+        self.data = data
+        self.data.is_find = False
         self.n_steps = n_steps
-        assert 0 <= n_steps <= 1000000
-
-        self.templates = templates or seeds(template_seed, n_steps)
-        assert len(self.templates) >= n_steps
-        self.record = list(record or ())
-
-    def __eq__(self, other):
-        return isinstance(other, StateMachineRunner) and (
-            self.parameter_seed == other.parameter_seed and
-            self.template_seed == other.template_seed and
-            self.n_steps == other.n_steps
-        )
-
-    def __ne__(self, other):
-        return not self.__eq__(other)
-
-    def __hash__(self):
-        return hash((
-            self.parameter_seed,
-            self.template_seed,
-            self.n_steps,
-        ))
-
-    def __trackas__(self):
-        return (
-            StateMachineRunner,
-            self.parameter_seed, self.template_seed,
-            self.n_steps,
-            [data[1] for data in self.record],
-        )
-
-    def __repr__(self):
-        return (
-            u'StateMachineRunner(%d/%d steps)' % (
-                len([t for t in self.record if t != TOMBSTONE]),
-                self.n_steps,
-            )
-        )
 
     def run(self, state_machine, print_steps=None):
         if print_steps is None:
             print_steps = current_verbosity() >= Verbosity.debug
+        self.data.hypothesis_runner = state_machine
 
+        stopping_value = 1 - 1.0 / (1 + self.n_steps * 0.5)
         try:
-            for i in hrange(self.n_steps):
-                strategy = state_machine.steps()
+            steps = 0
+            while True:
+                if steps == self.n_steps:
+                    stopping_value = 0
+                self.data.start_example()
+                if not cu.biased_coin(self.data, stopping_value):
+                    self.data.stop_example()
+                    break
 
-                template_set = False
-                if i < len(self.record):
-                    if self.record[i] is TOMBSTONE:
-                        continue
-                    _, data = self.record[i]
-                    data = list(data)
-                    for data_index in hrange(len(data) - 1, -1, -1):
-                        try:
-                            template = strategy.from_basic(data[data_index])
-                            template_set = True
-                            break
-                        except BadData:
-                            pass
-                    if template_set:
-                        data[data_index], data[-1] = (
-                            data[-1], data[data_index]
-                        )
-                else:
-                    data = []
-                if not template_set:
-                    parameter = strategy.draw_parameter(Random(
-                        self.parameter_seed
-                    ))
-                    template = strategy.draw_template(
-                        Random(self.templates[i]), parameter)
-                    data.append(strategy.to_basic(template))
-
-                new_record = (
-                    strategy, data,
-                )
-                if i < len(self.record):
-                    self.record[i] = new_record
-                else:
-                    self.record.append(new_record)
-
-                strategy.from_basic(self.record[i][1][-1])
-                value = strategy.reify(template)
-
-                if print_steps:
-                    state_machine.print_step(value)
-                state_machine.execute_step(value)
+                value = self.data.draw(state_machine.steps())
+                steps += 1
+                if steps <= self.n_steps:
+                    if print_steps:
+                        state_machine.print_step(value)
+                    state_machine.execute_step(value)
+                self.data.stop_example()
         finally:
             state_machine.teardown()
 
@@ -320,193 +244,33 @@ class StateMachineSearchStrategy(SearchStrategy):
     def __init__(self, settings=None):
         self.program_size = (settings or Settings.default).stateful_step_count
 
-    def __repr__(self):
-        return u'StateMachineSearchStrategy()'
-
-    def reify(self, template):
-        return template
-
-    def draw_parameter(self, random):
-        return (
-            random.getrandbits(64)
-        )
-
-    def draw_template(self, random, parameter_value):
-        parameter_seed = parameter_value
-        return StateMachineRunner(
-            parameter_seed,
-            random.getrandbits(64),
-            n_steps=self.program_size,
-        )
-
-    def to_basic(self, template):
-        return [
-            template.parameter_seed,
-            template.template_seed,
-            template.n_steps,
-            [
-                [data[1]]
-                if data != TOMBSTONE else None
-                for data in template.record
-            ]
-        ]
-
-    def from_basic(self, data):
-        check_data_type(list, data)
-        check_length(4, data)
-        check_data_type(integer_types, data[0])
-        check_data_type(integer_types, data[1])
-        check_data_type(integer_types, data[2])
-        check_data_type(list, data[3])
-
-        if data[2] < 0:
-            raise BadData(u'Invalid negative number of steps: %d' % (
-                data[2],
-            ))
-        if data[2] > Settings.default.stateful_step_count * 1000:
-            raise BadData(u'Implausibly large number of steps: %d' % (
-                data[2],
-            ))
-
-        record = []
-
-        for record_data in data[3]:
-            if record_data is None:
-                record.append(TOMBSTONE)
-            else:
-                check_data_type(list, record_data)
-                check_length(1, record_data)
-                record.append((None, record_data[0]))
-        return StateMachineRunner(
-            parameter_seed=data[0], template_seed=data[1],
-            n_steps=data[2],
-            record=record,
-        )
-
-    def simplifiers(self, random, template):
-        yield self.cut_steps
-        yield self.random_discards
-        yield self.delete_elements
-        for i in hrange(len(template.record)):
-            if template.record[i] != TOMBSTONE:
-                strategy, data = template.record[i]
-                if strategy is None:
-                    continue
-                child_template = strategy.from_basic(data[-1])
-                for simplifier in strategy.simplifiers(random, child_template):
-                    yield self.convert_simplifier(strategy, simplifier, i)
-
-    def convert_simplifier(self, strategy, simplifier, i):
-        def accept(random, template):
-            if i >= len(template.record):
-                return
-            if template.record[i][0] is not strategy:
-                return
-
-            reconstituted = strategy.from_basic(template.record[i][1][-1])
-
-            for t in simplifier(random, reconstituted):
-                new_record = list(template.record)
-                existing = new_record[i]
-                new_record[i] = (existing[0], list(existing[1]))
-                new_record[i][1][-1] = strategy.to_basic(t)
-                yield StateMachineRunner(
-                    parameter_seed=template.parameter_seed,
-                    template_seed=template.template_seed,
-                    templates=template.templates,
-                    n_steps=template.n_steps,
-                    record=new_record,
-                )
-        accept.__name__ = str(u'convert_simplifier(%s, %d)' % (
-            simplifier.__name__, i
-        ))
-        return accept
-
-    def random_discards(self, random, template):
-        live = len([
-            r for r in template.record if r != TOMBSTONE
-        ])
-        if live < 10:
-            return
-
-        for k in hrange(1, 8):
-            for _ in hrange(10):
-                new_record = list(template.record)
-                for i in hrange(len(template.record)):
-                    if new_record[i] != TOMBSTONE:
-                        if random.randint(0, 9) <= k:
-                            new_record[i] = TOMBSTONE
-                yield StateMachineRunner(
-                    parameter_seed=template.parameter_seed,
-                    template_seed=template.template_seed,
-                    templates=template.templates,
-                    n_steps=template.n_steps,
-                    record=new_record,
-                )
-
-    def cut_steps(self, random, template):
-        if len(template.record) < template.n_steps:
-            yield StateMachineRunner(
-                parameter_seed=template.parameter_seed,
-                template_seed=template.template_seed,
-                templates=template.templates,
-                n_steps=len(template.record),
-                record=template.record,
-            )
-        mid = 0
-        while True:
-            next_mid = (template.n_steps + mid) // 2
-            if next_mid == mid:
-                break
-            mid = next_mid
-            yield StateMachineRunner(
-                parameter_seed=template.parameter_seed,
-                template_seed=template.template_seed,
-                templates=template.templates,
-                n_steps=mid,
-                record=template.record,
-            )
-            new_record = list(template.record)
-            for i in hrange(min(mid, len(new_record))):
-                new_record[i] = TOMBSTONE
-            yield StateMachineRunner(
-                parameter_seed=template.parameter_seed,
-                template_seed=template.template_seed,
-                templates=template.templates,
-                n_steps=template.n_steps,
-                record=new_record,
-            )
-
-    def delete_elements(self, random, template):
-        deletes = 0
-        indices = list(hrange(len(template.record)))
-        random.shuffle(indices)
-        for i in indices:
-            if deletes >= 10:
-                break
-            if template.record[i] != TOMBSTONE:
-                deletes += 1
-                new_record = list(template.record)
-                new_record[i] = TOMBSTONE
-                yield StateMachineRunner(
-                    parameter_seed=template.parameter_seed,
-                    template_seed=template.template_seed,
-                    templates=template.templates,
-                    n_steps=template.n_steps,
-                    record=new_record,
-                )
+    def do_draw(self, data):
+        return StateMachineRunner(data, self.program_size)
 
 
 Rule = namedtuple(
     u'Rule',
     (u'targets', u'function', u'arguments', u'precondition',
      u'parent_rule')
-
 )
 
-Bundle = namedtuple(u'Bundle', (u'name',))
+self_strategy = runner()
 
 
+class Bundle(SearchStrategy):
+
+    def __init__(self, name):
+        self.name = name
+
+    def do_draw(self, data):
+        machine = data.draw(self_strategy)
+        bundle = machine.bundle(self.name)
+        if not bundle:
+            data.mark_invalid()
+        reference = bundle.pop()
+        bundle.insert(integer_range(data, 0, len(bundle)), reference)
+        return machine.names_to_values[reference.name]
+
 RULE_MARKER = u'hypothesis_stateful_rule'
 PRECONDITION_MARKER = u'hypothesis_stateful_precondition'
 
@@ -536,10 +300,8 @@ def rule(targets=(), target=None, **kwargs):
     def accept(f):
         parent_rule = getattr(f, RULE_MARKER, None)
         if parent_rule is not None:
-            note_deprecation(
-                'Applying the rule decorator to a function that is already '
-                'decorated by rule is deprecated. Please assign the result '
-                'of the rule function to separate names in your class.',
+            raise InvalidDefinition(
+                'A function cannot be used for two distinct rules. ',
                 Settings.default,
             )
         precondition = getattr(f, PRECONDITION_MARKER, None)
@@ -597,13 +359,7 @@ def precondition(precond):
     return decorator
 
 
-class SimpleSampledFromStrategy(SampledFromStrategy):
-
-    def draw_parameter(self, random):
-        return None
-
-    def draw_template(self, random, parameter_value):
-        return random.randint(0, len(self.elements) - 1)
+ShuffleBundle = namedtuple('ShuffleBundle', ('bundle', 'swaps'))
 
 
 class RuleBasedStateMachine(GenericStateMachine):
@@ -629,11 +385,23 @@ class RuleBasedStateMachine(GenericStateMachine):
         self.bundles = {}
         self.name_counter = 1
         self.names_to_values = {}
+        self.__stream = CUnicodeIO()
+        self.__printer = RepresentationPrinter(self.__stream)
+
+    def __pretty(self, value):
+        self.__stream.seek(0)
+        self.__stream.truncate(0)
+        self.__printer.output_width = 0
+        self.__printer.buffer_width = 0
+        self.__printer.buffer.clear()
+        self.__printer.pretty(value)
+        self.__printer.flush()
+        return self.__stream.getvalue()
 
     def __repr__(self):
         return u'%s(%s)' % (
             type(self).__name__,
-            repr(self.bundles),
+            nicerepr(self.bundles),
         )
 
     def upcoming_name(self):
@@ -670,8 +438,6 @@ class RuleBasedStateMachine(GenericStateMachine):
                     parent_rule=None):
         converted_arguments = {}
         for k, v in arguments.items():
-            if not isinstance(v, Bundle):
-                v = strategy(v)
             converted_arguments[k] = v
         if cls in cls._rules_per_class:
             target = cls._rules_per_class[cls]
@@ -692,34 +458,39 @@ class RuleBasedStateMachine(GenericStateMachine):
             valid = True
             if rule.precondition is not None and not rule.precondition(self):
                 continue
-            for k, v in rule.arguments.items():
+            for k, v in sorted(rule.arguments.items()):
                 if isinstance(v, Bundle):
                     bundle = self.bundle(v.name)
                     if not bundle:
                         valid = False
                         break
-                    else:
-                        v = SimpleSampledFromStrategy(bundle)
                 converted_arguments[k] = v
             if valid:
                 strategies.append(TupleStrategy((
-                    JustStrategy(rule),
+                    just(rule),
                     FixedKeysDictStrategy(converted_arguments)
                 ), tuple))
         if not strategies:
             raise InvalidDefinition(
                 u'No progress can be made from state %r' % (self,)
             )
-        return one_of_strategies(strategies)
+
+        for name, bundle in self.bundles.items():
+            if len(bundle) > 1:
+                strategies.append(
+                    builds(
+                        ShuffleBundle, just(name),
+                        lists(integers(0, len(bundle) - 1))))
+
+        return one_of(strategies)
 
     def print_step(self, step):
+        if isinstance(step, ShuffleBundle):
+            return
         rule, data = step
         data_repr = {}
         for k, v in data.items():
-            if isinstance(v, VarReference):
-                data_repr[k] = v.name
-            else:
-                data_repr[k] = repr(v)
+            data_repr[k] = self.__pretty(v)
         self.step_count = getattr(self, u'step_count', 0) + 1
         report(u'Step #%d: %s%s(%s)' % (
             self.step_count,
@@ -729,14 +500,19 @@ class RuleBasedStateMachine(GenericStateMachine):
         ))
 
     def execute_step(self, step):
+        if isinstance(step, ShuffleBundle):
+            bundle = self.bundle(step.bundle)
+            for i in step.swaps:
+                bundle.insert(i, bundle.pop())
+            return
         rule, data = step
         data = dict(data)
-        for k, v in data.items():
-            if isinstance(v, VarReference):
-                data[k] = self.names_to_values[v.name]
         result = rule.function(self, **data)
         if rule.targets:
             name = self.new_name()
             self.names_to_values[name] = result
+            self.__printer.singleton_pprinters.setdefault(
+                id(result), lambda obj, p, cycle: p.text(name),
+            )
             for target in rule.targets:
                 self.bundle(target).append(VarReference(name))
